/*
 * Part of the PiyoPiyoJ project.
 * (PiyoPiyo Java)
 *
 *        DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *                    Version 2, December 2004 
 *
 * Modified by gamemanj <asdd2808@gmail.com> to remove expletives
 * Original copyright (C) 2004 Sam Hocevar <sam@hocevar.net> 
 *
 * Everyone is permitted to copy and distribute verbatim or modified 
 * copies of this license document, and changing it is allowed as long 
 * as the name is changed. 
 *
 *            DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION 
 *
 *  0. You just DO WHATEVER YOU WANT TO.
 */
package piyopiyoj.games.ui;

import gabien.IPeripherals;
import gabien.ui.UIElement;
import gabien.ui.UILabel;
import gabien.ui.UISplitterLayout;
import gabien.ui.UITextButton;
import piyopiyoj.PlayerControls;

/**
 * Handles play control logic for a PiyoPiyoPlayer.
 */
public class UIPlayerControls extends UIElement.UIProxy {

    public final PlayerControls corePlayer;

    private UITextButton play = new UITextButton(">", 16, new Runnable() {
        @Override
        public void run() {
            playReverse.state = false;
            corePlayer.playing = play.state;
            corePlayer.reverse = false;
        }
    }).togglable(false);

    private UITextButton playReverse = new UITextButton("<", 16, new Runnable() {
        @Override
        public void run() {
            play.state = false;
            corePlayer.playing = playReverse.state;
            corePlayer.reverse = true;
        }
    }).togglable(false);

    private UITextButton halfSpeed = new UITextButton("/", 16, new Runnable() {
        @Override
        public void run() {
            corePlayer.halfSpeed = halfSpeed.state;
        }
    }).togglable(false);

    private UITextButton toStart = new UITextButton("<<", 16, new Runnable() {
        @Override
        public void run() {
            corePlayer.target.currentFrame = 0;
        }
    });
    private UITextButton toEnd = new UITextButton(">>", 16, new Runnable() {
        @Override
        public void run() {
            corePlayer.target.currentFrame = corePlayer.target.piyoFile.loopEnd;
        }
    });

    private UITextButton loop = new UITextButton("Loop", 16, new Runnable() {
        @Override
        public void run() {
            corePlayer.target.handleLoop = loop.state;
        }
    }).togglable(false);

    private UILabel songDetails = new UILabel("insert details", 16);

    public UIPlayerControls(PlayerControls myPlayer) {
        corePlayer = myPlayer;

        play.state = corePlayer.playing && (!corePlayer.reverse);
        playReverse.state = corePlayer.playing && corePlayer.reverse;
        halfSpeed.state = corePlayer.halfSpeed;
        loop.state = corePlayer.target.handleLoop;

        proxySetElement(UISplitterLayout.produceSideAlignedList(true, false,
            songDetails,
            loop,
            toStart,
            playReverse,
            halfSpeed,
            play,
            toEnd
        ), true);
    }

    @Override
    public void update(double deltaTime, boolean selected, IPeripherals peripherals) {
        double time = convertTime(corePlayer.target.currentFrame, corePlayer.target.withinBeatPosition);
        double timeEnd = convertTime(corePlayer.target.piyoFile.loopEnd, 0);
        songDetails.text = corePlayer.target.currentFrame + "/" + (corePlayer.target.piyoFile.musicFrames.length / 4) + " (" + time + "s/" + timeEnd + "s)";
        super.update(deltaTime, selected, peripherals);
    }

    private double convertTime(int frames, double offset) {
        double time = (frames * corePlayer.target.piyoFile.beatTime()) + offset;
        time *= 10;
        time = (int) time;
        time /= 10;
        return time;
    }
}
